import logging
import os
import shutil
import urllib
import urlparse

import tornado.web

logger = logging.getLogger(__name__)


class IndexHandler(tornado.web.RequestHandler):
    def initialize(self, core, config):
        self.core = core
        self.config = config
        self.application.settings["cookie_secret"] = self.config['auto']['cookie_secret']

    def get(self, path):
        return self.render('index.html', authed=self.get_secure_cookie('authed', None))

    def post(self, path):
        action = self.get_body_argument("action", False)
        if action:
            track = self.core.playback.get_current_track().get()
            if track:
                track_path = urllib.url2pathname(urlparse.urlparse(track.uri.split('file://')[1]).path)
                if action == 'delete-track':
                    self.core.playback.next()
                    try:
                        os.remove(track_path)
                        logger.info("Deleted track '{}'".format(track.name))
                    except OSError:
                        logger.error("Could not delete track '{}'".format(track.name), exc_info=1)
                elif action == 'delete-album':
                    album_path = os.path.dirname(track_path)
                    self.core.tracklist.clear()
                    try:
                        shutil.rmtree(album_path, True)
                        logger.info("Deleted album '{}'".format(track.album.name))
                    except shutil.Error:
                        logger.error("Could not delete album '{}'".format(track.album.name), exc_info=1)
                elif action == 'move-album':
                    print('Move current album')
                    return self.redirect('/move')
        elif self.get_body_argument('lock', False):
            self.clear_cookie('authed')
            return self.render('index.html', authed=False)
        elif self.get_body_argument("password") == self.config['auto']['admin_key']:
            self.set_secure_cookie('authed', 'true')
            return self.render('index.html', authed=True)

        return self.get(path)

    def get_template_path(self):
        return os.path.join(os.path.dirname(__file__), 'static')


class MoveHandler(tornado.web.RequestHandler):
    def initialize(self, core, config):
        self.core = core
        self.config = config

        # Store path of currently playing album
        track = self.core.playback.get_current_track().get()
        if track:
            track_path = urllib.url2pathname(urlparse.urlparse(track.uri.split('file://')[1]).path)
            self.album_path = os.path.dirname(track_path)
        else:
            return self.redirect('/auto')

        # Read all section config to an array. Ready for dynamic amount of sections
        self.sections = []
        section = 0

        exists = "s{}_folder".format(section) in self.config['auto']
        while exists:
            self.sections.append({
                "start": self.config['auto']["s{}_start".format(section)],
                "folder": self.config['auto']["s{}_folder".format(section)],
                "max_volume": self.config['auto']["s{}_max_volume".format(section)]
            })
            section += 1
            exists = "s{}_folder".format(section) in self.config['auto']

    def get(self):
        return self.render('static/move.html', sections=self.sections, current=self.album_path)

    def post(self):
        to = self.get_body_argument("to", False)

        if to:
            destination_uri = os.path.join(self.config['auto']['base_path'], to)
            destination_path = urllib.url2pathname(urlparse.urlparse(destination_uri.split('file://')[1]).path)

            logger.info("Moving '{}' to '{}'".format(self.album_path, destination_path))

            self.core.tracklist.clear()
            try:
                shutil.move(self.album_path, destination_path)
                logger.info("Moved album '{}' to '{}'".format(self.album_path, destination_path))
            except shutil.Error:
                logger.error("Could not move album '{}' to '{}'".format(self.album_path, destination_path), exc_info=1)
        else:
            logger.error('Destination not supplied, move canceled')

        return self.redirect('/auto')


class VolumeHandler(tornado.web.RequestHandler):
    def initialize(self, core):
        self.core = core

    def get(self):
        self.write("{}".format(80))



import datetime
import logging
import random
import sys
import urllib2

from mopidy import core

import pykka

import mopidy_auto

logger = logging.getLogger(__name__)


class AutoFrontend(pykka.ThreadingActor, core.CoreListener):
    def __init__(self, config, core):
        super(AutoFrontend, self).__init__()
        self.config = config
        self.core = core

        # One history per section
        self.history = {0: [], 1: [], 2: []}

        # Keep track of the current playing section
        self.section = None

        # Where sections folders are located
        self.base_path = self.config['auto']['base_path']

        # Maximum amount of tracks to add per album
        self.max_tracks = self.config['auto']['max_tracks']
        if self.max_tracks <= 0:
            self.max_tracks = sys.maxsize

        logger.info("Mopidy-Auto v%s", mopidy_auto.__version__)
        logger.info("Auto base path: %s, max tracks: %d", self.base_path, self.max_tracks)

        # Read all section config to an array. Ready for dynamic amount of sections
        self.sections = []
        section = 0

        exists = "s{}_folder".format(section) in self.config['auto']
        while exists:
            self.sections.append({
                "start": self.config['auto']["s{}_start".format(section)],
                "folder": self.config['auto']["s{}_folder".format(section)],
                "max_volume": self.config['auto']["s{}_max_volume".format(section)]
            })
            section += 1
            exists = "s{}_folder".format(section) in self.config['auto']

        if not self.sections:
            logger.error('Could not find any auto sections')
        else:
            logger.info('Found the following auto sections:')
            for section in self.sections:
                logger.info("Start: %s Folder: %s, Max Volume: %d%%",
                            section['start'],
                            section['folder'],
                            section['max_volume'])

    # Mopidy events
    # If tracklist has changed and is empty, play a random album
    def tracklist_changed(self):
        if self.core.tracklist.get_length().get() == 0:
            self.play_random_album()

    # If playback has stopped,
    def track_playback_ended(self, tl_track, time_position):
        # on the last track of the tracklist, play a random album
        if self.core.tracklist.index(tl_track).get() == self.core.tracklist.get_length().get() - 1:
            self.play_random_album()
        # and time is in new section, play random album
        elif self.section != self.get_section_by_time()[1]:
            logger.info("Time has changed to new section, play random album")
            self.play_random_album()

    # When resuming playback, check that time still in correct section, otherwise play new random album
    def track_playback_resumed(self, tl_track, time_position):
        if self.section != self.get_section_by_time()[1]:
            logger.info("Time has changed to new section, play random album")
            self.play_random_album()

    # Functions
    def play_random_album(self):
        section_index, self.section = self.get_section_by_time()
        logger.info("Auto play of random album, folder: %s", self.section['folder'])

        # Decrease volume if it's higher than sections max_volume
        if self.core.mixer.get_volume().get() > self.section['max_volume']:
            self.core.mixer.set_volume(self.section['max_volume'])

        # Find a random album from this sections URI
        uri = self.base_path + self.section['folder']
        tracks = self.get_random_album(uri, section_index)

        # If no tracks were found, try again
        while not tracks:
            tracks = self.get_random_album(uri, section_index)

        # Play the tracks
        self.play_uris(tracks)

    def get_section_by_time(self):
        now = datetime.datetime.now()
        now_minutes = now.hour * 60 + now.minute

        # Loop through sections in reverse to find the most accurate section
        for section in reversed(self.sections):
            section_start = section['start'].split(':')
            section_minutes = int(section_start[0]) * 60 + int(section_start[1])
            logger.debug('Get section by time: {} >= {}? {}'
                         .format(now_minutes, section_minutes, now_minutes >= section_minutes))
            if now_minutes >= section_minutes:
                return self.sections.index(section), section

        # If no section match, the first session starts after midnight
        # Use the last section that should play into the morning
        return self.sections.index(self.section[-1]), self.section[-1]

    def get_random_album(self, uri, section_index, root=True):
        track_uris = []
        logger.debug("Navigating file structure, URI: %s", uri)

        # Get all directories or tracks at the current URI
        refs = self.core.library.browse(uri).get()

        # Loop through all refs looking for tracks
        for ref in refs:
            if ref.type == 'track':
                track_uris.append(ref.uri)

                # Limit amount of tracks added
                if len(track_uris) >= self.max_tracks:
                    logger.info("Reached maximum tracks from same folder: %d", self.max_tracks)
                    break

        # If tracks were found, save album to history and return tracks
        if len(track_uris) > 0:
            self.history[section_index].append(uri)
            logger.info("Found %d tracks in folder '%s'", len(track_uris),
                        urllib2.unquote(uri.decode('ascii', 'ignore')))

            return track_uris

        # If not, limit refs to unplayed ones
        refs = self.get_unplayed_directories(refs, section_index, root)

        # Handle no folders returned
        if len(refs) <= 0:
            return None

        #  and recursively get a random one
        rand_idx = random.randint(0, len(refs) - 1)
        return self.get_random_album(refs[rand_idx].uri, section_index, root=False)

    def play_uris(self, uris):
        # Clear tracklist
        self.core.tracklist.clear()

        # Add all tracks by URIs
        self.core.tracklist.add(uris=uris)

        # Start playing
        self.core.playback.play()

    def get_unplayed_directories(self, refs, section_index, root):
        if len(refs) == 0:
            return []

        # Get all directories that are not in this sections history
        unplayed = [x for x in refs if x.uri not in self.history[section_index]]

        # If all albums on this level have been played:
        if len(unplayed) == 0:
            if root:
                # All folder on the root level has been played
                logger.info("Unique albums depleted. Clearing history for section %d", section_index)

                # Return a list containing all albums except the last one played
                unplayed = [x for x in refs if x.uri != self.history[section_index][-1]]

                # And clear the sections history
                self.history[section_index] = []
            else:
                # Sub level, add the parent of the refs to history
                parent_uri = refs[0].uri.rpartition('/')[0]
                self.history[section_index].append(parent_uri)

                # Return an empty list which will be handled as a empty leaf
                unplayed = []

        return unplayed



from __future__ import unicode_literals


from mopidy_auto import Extension


def test_get_default_config():
    ext = Extension()

    config = ext.get_default_config()

    assert '[auto]' in config
    assert 'enabled = true' in config


def test_get_config_schema():
    ext = Extension()

    schema = ext.get_config_schema()
    assert 'admin_key' in schema
    assert 'base_path' in schema
    assert 'max_tracks' in schema
    for index in range(3):
        assert "s{}_start".format(index) in schema
        assert "s{}_folder".format(index) in schema
        assert "s{}_max_volume".format(index) in schema

# TODO Write more test



from __future__ import unicode_literals

import logging
import os

from mopidy import config, ext

import tornado.web

from .web import IndexHandler, MoveHandler, VolumeHandler

__version__ = '0.5.0'

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = 'Mopidy-Auto'
    ext_name = 'auto'
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super(Extension, self).get_config_schema()
        schema['admin_key'] = config.String()
        schema['cookie_secret'] = config.String()
        schema['base_path'] = config.String()
        schema['max_tracks'] = config.Integer()
        for index in range(3):
            schema["s{}_start".format(index)] = config.String()
            schema["s{}_folder".format(index)] = config.String()
            schema["s{}_max_volume".format(index)] = config.Integer()
        return schema

    def setup(self, registry):
        from .frontend import AutoFrontend
        registry.add('frontend', AutoFrontend)

        registry.add('http:app', {
            'name': self.ext_name,
            'factory': self.webapp
        })

    def webapp(self, config, core):
        return [
            (r'/(index.html)?', IndexHandler, dict(core=core, config=config)),
            (r'/move', MoveHandler, dict(core=core, config=config)),
            (r'/max_volume', VolumeHandler, dict(core=core)),
            (r'/(.*)', tornado.web.StaticFileHandler, {'path': os.path.join(os.path.dirname(__file__), 'static')}),
        ]
